"""Definition of the Passthru Component."""


import numpy as np

from openmdao.core.explicitcomponent import ExplicitComponent
from ...utils.misc import _none_or_unspecified
from ..._options import options as dymos_options


class ParameterComp(ExplicitComponent):
    """
    A component which simply passes a parameter input to an equivalent output.

    Parameters
    ----------
    time_options : TimeOptionsDictionary or None
        If None, specify time options for the creation of t_initial and t_duration inputs and outputs.
    **kwargs : dict
        Arguments to be passed to the component initialization method.

    Attributes
    ----------
    _vars : dict
        Container mapping name of variables to be muxed with additional data.
    _input_names : dict
        Container mapping name of variables to be muxed with associated inputs.
    """

    def __init__(self, time_options=None, **kwargs):
        """
        Instantiate MuxComp and populate private members.
        """
        super().__init__(**kwargs)
        self.time_options = time_options

        self._no_check_partials = not dymos_options['include_check_partials']

    def setup(self):
        """
        Add time-related I/O to the ParameterComp at setup, if provided.
        """
        time_options = self.time_options

        if time_options:
            ti_val = self.time_options['initial_val']
            td_val = self.time_options['duration_val']
            units = self.time_options['units']

            self.add_input(name='t_initial', val=ti_val, shape=(1,), units=units)
            self.add_output(name='t_initial_val', val=ti_val, shape=(1,), units=units)
            self.declare_partials(of='t_initial_val', wrt='t_initial', val=1.0)

            self.add_input(name='t_duration', val=td_val, shape=(1,), units=units)
            self.add_output(name='t_duration_val', val=td_val, shape=(1,), units=units)
            self.declare_partials(of='t_duration_val', wrt='t_duration', val=1.0)

    def add_parameter(self, name, val=1.0, shape=None, output_name=None,
                      units=None, desc='', tags=None, input_tags=None, output_tags=None, input_shape_by_conn=False,
                      input_copy_shape=None, output_shape_by_conn=False, output_copy_shape=None,
                      distributed=None, res_units=None, lower=None, upper=None, ref=1.0, ref0=0.0, res_ref=1.0,):
        """
        Add an input/output pair for a variable to this component.

        Parameters
        ----------
        name : str
            Name of the input variable in this component's namespace.
        val : float or list or tuple or ndarray or Iterable
            The initial value of the variable being added in user-defined units.
            Default is 1.0.
        shape : int or tuple or list or None
            Shape of this variable, only required if src_indices not provided and
            val is not an array. Default is None.
        output_name : str
            Name given to the output variale in this component's namespace.  If None, f'{name}_value' is used.
        units : str or None
            Units in which this input variable will be provided to the component
            during execution. Default is None, which means it is unitless.
        desc : str
            Description of the variable.
        tags : str or list of strs
            User defined tags that can be used to filter what gets listed when calling
            list_inputs and list_outputs.
        input_tags : str or list of strs
            User defined tags applied only to the inputs.
        output_tags : str or list of strs
            User defined tags applied only to the outputs.
        input_shape_by_conn : bool
            If True, shape this input to match its connected output.
        input_copy_shape : str or None
            If a str, that str is the name of a variable. Shape this input to match that of
            the named variable.
        output_shape_by_conn : bool
            If True, shape this output to match its connected input(s).
        output_copy_shape : str or None
            If a str, that str is the name of a variable. Shape this output to match that of
            the named variable.
        distributed : bool
            If True, this variable is a distributed variable, so it can have different sizes/values
            across MPI processes.
        res_units : str or None
            Units in which the residuals of this output will be given to the user when requested.
            Default is None, which means it has no units.
        lower : float or list or tuple or ndarray or Iterable or None
            Lower bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no lower bound.
            Default is None.
        upper : float or list or tuple or ndarray or or Iterable None
            Upper bound(s) in user-defined units. It can be (1) a float, (2) an array_like
            consistent with the shape arg (if given), or (3) an array_like matching the shape of
            val, if val is array_like. A value of None means this output has no upper bound.
            Default is None.
        ref : float or ndarray
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 1. Default is 1.
        ref0 : float or ndarray
            Scaling parameter. The value in the user-defined units of this output variable when
            the scaled value is 0. Default is 0.
        res_ref : float or ndarray
            Scaling parameter. The value in the user-defined res_units of this output's residual
            when the scaled value is 1. Default is 1.

        Returns
        -------
        input_meta : dict
            The metadata associated with the input.
        output_meta : dict
            The metadata associated with the output.
        """
        _out_name = output_name if output_name is not None else f'parameter_vals:{name}'

        _val = np.asarray(val)

        if shape in _none_or_unspecified:
            _shape = (1,)
            size = _val.size
        else:
            _shape = shape
            size = np.prod(shape)
        _out_shape = (1,) + tuple(shape)
        ar = np.arange(size, dtype=int)

        if tags is None:
            tags = []
        elif isinstance(tags, str):
            tags = [tags]

        if input_tags is None:
            input_tags = []
        elif isinstance(input_tags, str):
            input_tags = [input_tags]

        if output_tags is None:
            output_tags = []
        elif isinstance(output_tags, str):
            output_tags = [output_tags]

        in_val = _val
        out_val = np.expand_dims(in_val, axis=0)

        i_meta = self.add_input(name=f'parameters:{name}', val=in_val, shape=_shape, units=units, desc=desc,
                                tags=tags + input_tags,
                                shape_by_conn=input_shape_by_conn, copy_shape=input_copy_shape, distributed=distributed)

        o_meta = self.add_output(name=_out_name, val=out_val, shape=_out_shape, units=units,
                                 desc=desc, tags=tags + output_tags, res_units=res_units, ref=ref, ref0=ref0,
                                 res_ref=res_ref, lower=lower, upper=upper, shape_by_conn=output_shape_by_conn,
                                 copy_shape=output_copy_shape, distributed=distributed)

        self.declare_partials(of=_out_name, wrt=f'parameters:{name}', rows=ar, cols=ar, val=1.0)

        return i_meta, o_meta

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Mux the inputs into the appropriate outputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        outputs : Vector
            Unscaled, dimensional output variables read via outputs[key].
        discrete_inputs : Vector
            Discrete input variables read via discrete_inputs[key].
        discrete_outputs : Vector
            Discrete output variables read via discrete_outputs[key].
        """
        outputs.set_val(inputs.asarray())



import unittest

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal, assert_check_partials
from openmdao.utils.testing_utils import use_tempdirs

from dymos.examples.cart_pole.cartpole_dynamics import CartPoleDynamics


@use_tempdirs
class TestCartPoleDynamics(unittest.TestCase):
    def test_cartpole_ode(self):

        p = om.Problem()
        p.model.add_subsystem("dynamics", CartPoleDynamics(num_nodes=2), promotes=["*"])
        # set input values
        p.model.set_input_defaults("m_cart", 1.0, units="kg")
        p.model.set_input_defaults("m_pole", 0.3, units="kg")
        p.model.set_input_defaults("l_pole", 0.5, units="m")
        p.model.set_input_defaults("theta", [0.0, 1.0], units="rad")
        p.model.set_input_defaults("theta_dot", [0.0, 0.1], units="rad/s")
        p.model.set_input_defaults("f", [10, -10], units="N")

        # run model
        p.setup(check=False)
        p.run_model()
        # get outputs
        x_dotdot = p.get_val("x_dotdot", units="m/s**2")
        theta_dotdot = p.get_val("theta_dotdot", units="rad/s**2")
        assert_near_equal(x_dotdot, [10.0, -7.14331021], tolerance=1e-6)
        assert_near_equal(theta_dotdot, [-20.0, -8.79056677], tolerance=1e-6)

        # check partials
        partials = p.check_partials(compact_print=True, method="cs")
        assert_check_partials(partials, atol=1e-5, rtol=1e-5)


if __name__ == "___main__":
    unittest.main()



import unittest

import openmdao.api as om
import openmdao.func_api as omf
from dymos.utils.misc import om_version
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse
import dymos as dm
import numpy as np

from dymos.examples.balanced_field.balanced_field_length import make_balanced_field_length_problem

try:
    import jax
except ImportError:
    jax = None


def runway_ode(rho, S, CD0, CL0, CL_max, alpha_max, h_w, AR, e, span, T, mu_r, m, v, h, alpha):
    g = 9.80665
    W = m * g
    v_stall = np.sqrt(2 * W / rho / S / CL_max)
    v_over_v_stall = v / v_stall

    CL = CL0 + (alpha / alpha_max) * (CL_max - CL0)
    K_nom = 1.0 / (np.pi * AR * e)
    b = span / 2.0
    fact = ((h + h_w) / b) ** 1.5
    K = K_nom * 33 * fact / (1.0 + 33 * fact)

    q = 0.5 * rho * v ** 2
    L = q * S * CL
    D = q * S * (CD0 + K * CL ** 2)

    # Compute the downward force on the landing gear
    calpha = np.cos(alpha)
    salpha = np.sin(alpha)

    # Runway normal force
    F_r = m * g - L * calpha - T * salpha

    # Compute the dynamics
    v_dot = (T * calpha - D - F_r * mu_r) / m
    r_dot = v

    return CL, q, L, D, K, F_r, v_dot, r_dot, W, v_stall, v_over_v_stall


def climb_ode(rho, S, CD0, CL0, CL_max, alpha_max, h_w, AR, e, span, T, m, v, h, alpha, gam):
    g = 9.80665
    W = m * g
    v_stall = np.sqrt(2 * W / rho / S / CL_max)
    v_over_v_stall = v / v_stall

    CL = CL0 + (alpha / alpha_max) * (CL_max - CL0)
    K_nom = 1.0 / (np.pi * AR * e)
    b = span / 2.0
    fact = ((h + h_w) / b) ** 1.5
    K = K_nom * 33 * fact / (1.0 + 33 * fact)

    q = 0.5 * rho * v ** 2
    L = q * S * CL
    D = q * S * (CD0 + K * CL ** 2)

    # Compute the downward force on the landing gear
    calpha = np.cos(alpha)
    salpha = np.sin(alpha)

    # Runway normal force
    F_r = m * g - L * calpha - T * salpha

    # Compute the dynamics
    cgam = np.cos(gam)
    sgam = np.sin(gam)
    v_dot = (T * calpha - D) / m - g * sgam
    gam_dot = (T * salpha + L) / (m * v) - (g / v) * cgam
    h_dot = v * sgam
    r_dot = v * cgam

    return CL, q, L, D, K, F_r, v_dot, r_dot, W, v_stall, v_over_v_stall, gam_dot, h_dot


def wrap_ode_func(num_nodes, mode, grad_method='jax', jax_jit=True):
    """
    Returns the metadata from omf needed to create a new ExplciitFuncComp.
    """
    nn = num_nodes
    ode_func = runway_ode if mode == 'runway' else climb_ode

    meta = (omf.wrap(ode_func)
            .add_input('rho', val=1.225, desc='atmospheric density at runway', units='kg/m**3')
            .add_input('S', val=124.7, desc='aerodynamic reference area', units='m**2')
            .add_input('CD0', val=0.03, desc='zero-lift drag coefficient', units=None)
            .add_input('CL0', val=0.5, desc='zero-alpha lift coefficient', units=None)
            .add_input('CL_max', val=2.0, desc='maximum lift coefficient for linear fit', units=None)
            .add_input('alpha_max', val=np.radians(10), desc='angle of attack at CL_max', units='rad')
            .add_input('h_w', val=1.0, desc='height of the wing above the CG', units='m')
            .add_input('AR', val=9.45, desc='wing aspect ratio', units=None)
            .add_input('e', val=0.801, desc='Oswald span efficiency factor', units=None)
            .add_input('span', val=35.7, desc='Wingspan', units='m')
            .add_input('T', val=1.0, desc='thrust', units='N')

            # Dynamic inputs (can assume a different value at every node)
            .add_input('m', shape=(nn,), desc='aircraft mass', units='kg')
            .add_input('v', shape=(nn,), desc='aircraft true airspeed', units='m/s')
            .add_input('h', shape=(nn,), desc='altitude', units='m')
            .add_input('alpha', shape=(nn,), desc='angle of attack', units='rad')

            # Outputs
            .add_output('CL', shape=(nn,), desc='lift coefficient', units=None)
            .add_output('q', shape=(nn,), desc='dynamic pressure', units='Pa')
            .add_output('L', shape=(nn,), desc='lift force', units='N')
            .add_output('D', shape=(nn,), desc='drag force', units='N')
            .add_output('K', val=np.ones(nn), desc='drag-due-to-lift factor', units=None)
            .add_output('F_r', shape=(nn,), desc='runway normal force', units='N')
            .add_output('v_dot', shape=(nn,), desc='rate of change of speed', units='m/s**2',
                        tags=['dymos.state_rate_source:v'])
            .add_output('r_dot', shape=(nn,), desc='rate of change of range', units='m/s',
                        tags=['dymos.state_rate_source:r'])
            .add_output('W', shape=(nn,), desc='aircraft weight', units='N')
            .add_output('v_stall', shape=(nn,), desc='stall speed', units='m/s')
            .add_output('v_over_v_stall', shape=(nn,), desc='stall speed ratio', units=None)
            )

    if mode == 'runway':
        meta.add_input('mu_r', val=0.05, desc='runway friction coefficient', units=None)
    else:
        meta.add_input('gam', shape=(nn,), desc='flight path angle', units='rad')
        meta.add_output('gam_dot', shape=(nn,), desc='rate of change of flight path angle',
                        units='rad/s', tags=['dymos.state_rate_source:gam'])
        meta.add_output('h_dot', shape=(nn,), desc='rate of change of altitude', units='m/s',
                        tags=['dymos.state_rate_source:h'])

    meta.declare_coloring('*', method=grad_method)
    meta.declare_partials(of='*', wrt='*', method=grad_method)

    if om_version()[0] > (3, 35, 0):
        return om.ExplicitFuncComp(meta, derivs_method=grad_method, use_jit=jax_jit)
    else:
        return om.ExplicitFuncComp(meta, use_jax=grad_method == 'jax', use_jit=jax_jit)


@use_tempdirs
class TestBalancedFieldFuncComp(unittest.TestCase):

    @unittest.skipIf(jax is None, 'requires jax and jaxlib')
    @require_pyoptsparse('IPOPT')
    def test_balanced_field_func_comp_radau(self):
        p = make_balanced_field_length_problem(ode_class=wrap_ode_func,
                                               tx=dm.Radau(num_segments=3))

        dm.run_problem(p, run_driver=True, simulate=True)

        assert_near_equal(p.get_val('traj.rto.states:r')[-1], 2197.7, tolerance=0.01)

    @unittest.skipIf(jax is None, 'requires jax and jaxlib')
    @require_pyoptsparse('IPOPT')
    def test_balanced_field_func_comp_gl(self):
        p = make_balanced_field_length_problem(ode_class=wrap_ode_func,
                                               tx=dm.GaussLobatto(num_segments=3))

        dm.run_problem(p, run_driver=True, simulate=True)

        assert_near_equal(p.get_val('traj.rto.states:r')[-1], 2197.7, tolerance=0.01)



import openmdao.api as om
from .dynamic_pressure_comp import DynamicPressureComp
from .lift_drag_force_comp import LiftDragForceComp
from .cd0_comp import CD0Comp
from .kappa_comp import KappaComp
from .cla_comp import CLaComp
from .cl_comp import CLComp
from .cd_comp import CDComp
from .mach_comp import MachComp


class AeroGroup(om.Group):
    """
    The purpose of the AeroGroup is to compute the aerodynamic forces on the
    aircraft in the body frame.

    Parameters
    ----------
    v : float
        air-relative velocity (m/s)
    sos : float
        local speed of sound (m/s)
    rho : float
        atmospheric density (kg/m**3)
    alpha : float
        angle of attack (rad)
    S : float
        aerodynamic reference area (m**2)

    """
    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='mach_comp',
                           subsys=MachComp(num_nodes=nn),
                           promotes_inputs=['v', 'sos'],
                           promotes_outputs=['mach'])

        self.add_subsystem(name='cd0_comp',
                           subsys=CD0Comp(num_nodes=nn),
                           promotes_inputs=['mach'],
                           promotes_outputs=['CD0'])

        self.add_subsystem(name='kappa_comp',
                           subsys=KappaComp(num_nodes=nn),
                           promotes_inputs=['mach'],
                           promotes_outputs=['kappa'])

        self.add_subsystem(name='cla_comp',
                           subsys=CLaComp(num_nodes=nn),
                           promotes_inputs=['mach'],
                           promotes_outputs=['CLa'])

        self.add_subsystem(name='CL_comp',
                           subsys=CLComp(num_nodes=nn),
                           promotes_inputs=['alpha', 'CLa'],
                           promotes_outputs=['CL'])

        self.add_subsystem(name='CD_comp',
                           subsys=CDComp(num_nodes=nn),
                           promotes_inputs=['CD0', 'alpha', 'CLa', 'kappa'],
                           promotes_outputs=['CD'])

        self.add_subsystem(name='q_comp',
                           subsys=DynamicPressureComp(num_nodes=nn),
                           promotes_inputs=['rho', 'v'],
                           promotes_outputs=['q'])

        self.add_subsystem(name='lift_drag_force_comp',
                           subsys=LiftDragForceComp(num_nodes=nn),
                           promotes_inputs=['CL', 'CD', 'q', 'S'],
                           promotes_outputs=['f_lift', 'f_drag'])



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs

import dymos as dm
from dymos.examples.brachistochrone.brachistochrone_vector_states_ode \
    import BrachistochroneVectorStatesODE


@use_tempdirs
class TestBrachistochroneVectorPathConstraints(unittest.TestCase):

    def test_brachistochrone_vector_state_path_constraints_radau_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta_rate', loc='final', equals=0.0)
        phase.add_boundary_constraint('theta_rate2', loc='final', equals=0.0)
        phase.add_path_constraint('pos', indices=[1], lower=5)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos')[:, 1]), 5.0,
                          tolerance=1.0E-3)

    def test_brachistochrone_vector_ode_path_constraints_radau_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', indices=[1],
                                  lower=-4, upper=4)

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)

    def test_brachistochrone_vector_ode_path_constraints_radau_no_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.Radau(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', lower=-4, upper=12)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)

    def test_brachistochrone_vector_state_path_constraints_gl_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg', rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos', indices=[1], lower=5)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos')[:, 1]),
                          5,
                          tolerance=1.0E-2)

    def test_brachistochrone_vector_ode_path_constraints_gl_partial_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg',
                          rate_continuity=True, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_boundary_constraint('theta_rate', loc='final', equals=0.0)
        phase.add_boundary_constraint('theta_rate2', loc='final', equals=0.0)
        phase.add_path_constraint('pos_dot', indices=[1],
                                  lower=-4, upper=4)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, 1]), -4.0,
                          tolerance=1.0E-3)

    def test_brachistochrone_vector_ode_path_constraints_gl_no_indices(self):

        p = om.Problem(model=om.Group())

        p.driver = om.ScipyOptimizeDriver()
        p.driver.declare_coloring()

        phase = dm.Phase(ode_class=BrachistochroneVectorStatesODE,
                         transcription=dm.GaussLobatto(num_segments=20, order=3))

        p.model.add_subsystem('phase0', phase)

        phase.set_time_options(fix_initial=True, duration_bounds=(.5, 10))

        phase.add_state('pos', fix_initial=True, fix_final=True)
        phase.add_state('v', fix_initial=True, fix_final=False)

        phase.add_control('theta', units='deg',
                          rate_continuity=False, lower=0.01, upper=179.9)

        phase.add_parameter('g', units='m/s**2', opt=False, val=9.80665)

        phase.add_path_constraint('pos_dot', lower=-4, upper=12)

        phase.add_timeseries_output('pos_dot')

        # Minimize time at the end of the phase
        phase.add_objective('time', loc='final', scaler=10)

        p.model.linear_solver = om.DirectSolver()
        p.setup(check=True, force_alloc_complex=True)

        phase.set_time_val(initial=0.0, duration=1.8016)

        pos0 = [0, 10]
        posf = [10, 5]

        phase.set_state_val('pos', [pos0, posf])
        phase.set_state_val('v', [0, 9.9])
        phase.set_control_val('theta', [5, 100])
        phase.set_parameter_val('g', 9.80665)

        p.run_driver()

        assert_near_equal(np.min(p.get_val('phase0.timeseries.pos_dot')[:, -1]),
                          -4,
                          tolerance=1.0E-2)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np
import openmdao.api as om

from ..phase.options import TimeOptionsDictionary
from ..utils.misc import get_rate_units
from ..utils.introspection import get_targets, _get_targets_metadata
from ..utils.ode_utils import _make_ode_system
from ..transcriptions.grid_data import GridData


class GridRefinementODESystem(om.Group):
    """
    Defines a group that performs grid refinement on an ODE.

    The Grid Refinement algorithms in Dymos use the following approach for computing
    errors in the transcription:

    1. The phase segmentation is reproduced, except each segment is of one polynomial order higher
    (for Radau) or two orders higher (for GaussLobatto).
    2. An interpolation matrix (L) and an integration matrix (I) is developed that takes the solution
    from 'all' nodes of the existing solution and interpolates it onto 'all' nodes of the higher-order segmentation.
    For states, this provides a reference set of interpolated values (x_hat).
    3. The ODE is evaluated at all nodes of the new, higher-order grid.  The resulting state-rate
    output values are used to integrate the state values onto 'all' nodes in the higher-order segmentation.
    These states are a new estimate of the state values at each node (x_prime).
    4. The error is computed by comparing x_hat to x_prime.

    This system is used to evaluate the ODE at all nodes within the higher-order segmentation.

    Parameters
    ----------
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def initialize(self):
        """
        Declare options for this Group.
        """
        self.options.declare('grid_data', types=GridData, desc='Container object for grid info')

        self.options.declare('time', types=TimeOptionsDictionary,
                             desc='Time options for the phase')

        self.options.declare('states', types=dict,
                             desc='Dictionary of state names/options for the segments parent Phase')

        self.options.declare('controls', default=None, types=dict, allow_none=True,
                             desc='Dictionary of control names/options for the segments parent Phase.')

        self.options.declare('parameters', default=None, types=dict, allow_none=True,
                             desc='Dictionary of parameter names/options for the segments '
                                  'parent Phase.')

        self.options.declare('ode_class',
                             desc='System defining the ODE')

        self.options.declare('ode_init_kwargs', types=dict, default={},
                             desc='Keyword arguments provided when initializing the ODE System')

        self.options.declare('calc_exprs', types=dict, default={},
                             desc='Calculation expressions from the parent phase.')

    def setup(self):
        """
        Add the ODE subsystem.
        """
        grid_data = self.options['grid_data']
        num_nodes = grid_data.num_nodes
        ode_class = self.options['ode_class']
        ode_init_kwargs = self.options['ode_init_kwargs']
        calc_exprs = self.options['calc_exprs']

        # The ODE System
        if ode_class is not None:
            ode_sys = _make_ode_system(ode_class=ode_class,
                                       num_nodes=num_nodes,
                                       ode_init_kwargs=ode_init_kwargs,
                                       calc_exprs=calc_exprs,
                                       parameter_options=self.options['parameters'])
            self.add_subsystem('ode', subsys=ode_sys)

    def configure(self):
        """
        Promote variables from the ODE.
        """
        grid_data = self.options['grid_data']
        num_nodes = grid_data.num_nodes

        # Configure time
        options = self.options['time']

        # time
        targets = get_targets(self.ode, 'time', options['targets'])
        for tgt in targets:
            self.promotes('ode', inputs=[(tgt, 'time')])
        if targets:
            self.set_input_defaults(name='time', val=np.ones(num_nodes), units=options['units'])

        # time_phase
        targets = get_targets(self.ode, 't_phase', options['time_phase_targets'])
        for tgt in targets:
            self.promotes('ode', inputs=[(tgt, 't_phase')])
        if targets:
            self.set_input_defaults(name='t_phase', val=np.ones(num_nodes), units=options['units'])

        # t_initial
        targets = get_targets(self.ode, 't_initial', options['t_initial_targets'])
        for tgt in targets:
            self.promotes('ode', inputs=[(tgt, 't_initial')])
        if targets:
            self.set_input_defaults(name='t_initial', val=np.ones(num_nodes), units=options['units'])

        # t_duration
        targets = get_targets(self.ode, 't_duration', options['t_duration_targets'])
        for tgt in targets:
            self.promotes('ode', inputs=[(tgt, 't_duration')])
        if targets:
            self.set_input_defaults(name='t_duration', val=np.ones(num_nodes), units=options['units'])

        # Configure the states
        for name, options in self.options['states'].items():
            targets = get_targets(self.ode, name, options['targets'])
            for tgt in targets:
                self.promotes('ode', inputs=[(tgt, f'states:{name}')])
            if targets:
                self.set_input_defaults(name=f'states:{name}',
                                        val=np.ones(num_nodes),
                                        units=options['units'])

        # Configure the controls
        for name, options in self.options['controls'].items():
            rate_units = get_rate_units(units=options['units'],
                                        time_units=self.options['time']['units'])
            rate2_units = get_rate_units(units=options['units'],
                                         time_units=self.options['time']['units'],
                                         deriv=2)

            targets = get_targets(self.ode, name, options['targets'])
            if targets:
                for tgt in targets:
                    self.promotes('ode', inputs=[(tgt, f'controls:{name}')])
                self.set_input_defaults(name=f'controls:{name}',
                                        val=np.ones(num_nodes),
                                        units=options['units'])

            targets = get_targets(self.ode, f'{name}_rate', options['rate_targets'])
            if targets:
                for tgt in targets:
                    self.promotes('ode', inputs=[(tgt, f'control_rates:{name}_rate')])
                self.set_input_defaults(name=f'control_rates:{name}_rate',
                                        val=np.ones(num_nodes),
                                        units=rate_units)

            targets = get_targets(self.ode, f'{name}_rate2', options['rate2_targets'])
            if targets:
                for tgt in targets:
                    self.promotes('ode', inputs=[(tgt, f'control_rates:{name}_rate2')])
                self.set_input_defaults(name=f'control_rates:{name}_rate2',
                                        val=np.ones(num_nodes),
                                        units=rate2_units)

        # Configure the parameters
        for name, options in self.options['parameters'].items():
            static_targets = options['static_targets']
            shape = options['shape']
            prom_name = f'parameters:{name}'
            targets = _get_targets_metadata(self.ode, name, options['targets'])
            for tgt, meta in targets.items():
                if tgt in static_targets:
                    self.promotes('ode', inputs=[(tgt, prom_name)])
                else:
                    self.promotes('ode', inputs=[(tgt, prom_name)],
                                  src_indices=om.slicer[np.zeros(num_nodes, dtype=int), ...])
            if targets:
                self.set_input_defaults(name=prom_name,
                                        src_shape=shape,
                                        units=options['units'])



import numpy as np
import openmdao.api as om


class Aerodynamics(om.ExplicitComponent):
    """
    Defines the aerodynamics for the shuttle reentry problem.

    References
    ----------
    .. [1] Betts, John T., Practical Methods for Optimal Control and Estimation Using Nonlinear
           Programming, p. 248, 2010.
    """

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('alpha', val=np.ones(nn), desc='angle of attack', units='deg')
        self.add_input('v', val=np.ones(nn), desc='velocity of shuttle', units='ft/s')
        self.add_input('rho', val=np.ones(nn), desc='local atmospheric density',
                       units='slug/ft**3')

        self.add_output('drag', val=np.ones(nn), desc='drag on shuttle', units='lb')
        self.add_output('lift', val=np.ones(nn), desc='lift on shuttle', units='lb')

        partial_range = np.arange(nn, dtype=int)

        self.declare_partials('drag', 'alpha', rows=partial_range, cols=partial_range)
        self.declare_partials('drag', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('drag', 'rho', rows=partial_range, cols=partial_range)
        self.declare_partials('lift', 'alpha', rows=partial_range, cols=partial_range)
        self.declare_partials('lift', 'v', rows=partial_range, cols=partial_range)
        self.declare_partials('lift', 'rho', rows=partial_range, cols=partial_range)

    def compute(self, inputs, outputs):
        a_0 = -.20704
        a_1 = .029244
        b_0 = .07854
        b_1 = -.61592e-2
        b_2 = .621408e-3
        S = 2690
        alpha = inputs['alpha']
        v = inputs['v']
        rho = inputs['rho']
        c_L = a_0 + a_1 * alpha
        c_D = b_0 + b_1 * alpha + b_2 * alpha ** 2

        outputs['drag'] = .5 * c_D * S * rho * v ** 2
        outputs['lift'] = .5 * c_L * S * rho * v ** 2

    def compute_partials(self, inputs, J):
        alpha = inputs['alpha']
        v = inputs['v']
        rho = inputs['rho']
        a_0 = -.20704
        a_1 = .029244
        b_0 = .07854
        b_1 = -.61592e-2
        b_2 = .621408e-3
        S = 2690
        c_L = a_0 + a_1 * alpha
        c_D = b_0 + b_1 * alpha + b_2 * alpha ** 2

        dD_dCD = .5 * S * rho * v ** 2
        dCD_dalpha = b_1 + 2 * alpha * b_2
        dL_dCL = dD_dCD
        dCL_dalpha = a_1

        J['drag', 'alpha'] = dD_dCD * dCD_dalpha
        J['drag', 'v'] = c_D * S * rho * v
        J['drag', 'rho'] = .5 * c_D * S * v ** 2
        J['lift', 'alpha'] = dL_dCL * dCL_dalpha
        J['lift', 'v'] = c_L * S * rho * v
        J['lift', 'rho'] = .5 * c_L * S * v ** 2



import unittest

import numpy as np

import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
import dymos as dm

from dymos.transcriptions.explicit_shooting.tau_comp import TauComp


class TestTauComp(unittest.TestCase):

    def test_eval(self):
        grid_data = dm.transcriptions.grid_data.GridData(num_segments=3, transcription='gauss-lobatto',
                                                         transcription_order=3)

        p = om.Problem()
        tau_comp = p.model.add_subsystem('tau_comp', TauComp(grid_data=grid_data, time_units='s'))
        p.setup(force_alloc_complex=True)

        tau_comp.options['segment_index'] = 0

        p.set_val('tau_comp.t_initial', 0)
        p.set_val('tau_comp.t_duration', 15)

        p.set_val('tau_comp.t', 1.5)
        p.run_model()

        frac = 0.1
        ptau_expected = -1 + 2 * frac
        assert_near_equal(p.get_val('tau_comp.ptau'), ptau_expected)

        frac = 1.5/5.0
        stau_expected = -1 + 2 * frac
        assert_near_equal(p.get_val('tau_comp.stau'), stau_expected)

        cpd = p.check_partials(method='cs')
        assert_check_partials(cpd)

    def test_eval_vectorized(self):
        grid_data = dm.transcriptions.grid_data.GridData(num_segments=3, transcription='gauss-lobatto',
                                                         transcription_order=3)

        p = om.Problem()
        tau_comp = p.model.add_subsystem('tau_comp', TauComp(vec_size=4, grid_data=grid_data, time_units='s'))
        p.setup(force_alloc_complex=True)

        tau_comp.options['segment_index'] = 0

        p.set_val('tau_comp.t_initial', 0)
        p.set_val('tau_comp.t_duration', 15)

        p.set_val('tau_comp.t', np.linspace(0, 5, 4))
        p.run_model()

        print(p.get_val('tau_comp.ptau'))

        cpd = p.check_partials(method='cs')
        assert_check_partials(cpd)



import numpy as np

import openmdao.api as om

from ..aero.aerodynamics_group import AerodynamicsGroup
from .lift_equilibrium_comp import LiftEquilibriumComp
from .thrust_equilibrium_comp import ThrustEquilibriumComp


class SteadyFlightEquilibriumGroup(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem('aero',
                           subsys=AerodynamicsGroup(num_nodes=nn),
                           promotes_inputs=['alt'])

        self.add_subsystem('thrust_eq_comp',
                           subsys=ThrustEquilibriumComp(num_nodes=nn),
                           promotes_inputs=['q', 'S', 'gam', 'alpha', 'W_total'],
                           promotes_outputs=['CT'])

        self.add_subsystem('lift_eq_comp',
                           subsys=LiftEquilibriumComp(num_nodes=nn),
                           promotes_inputs=['q', 'S', 'gam', 'alpha', 'W_total', 'CT'],
                           promotes_outputs=['CL_eq'])

        bal = self.add_subsystem(name='alpha_eta_balance',
                                 subsys=om.BalanceComp(),
                                 promotes_outputs=['alpha', 'eta'])

        self.connect('alpha', ('aero.alpha'))
        self.connect('eta', ('aero.eta'))

        bal.add_balance('alpha', units='rad', eq_units=None, lhs_name='CL_eq',
                        rhs_name='CL', val=0.01*np.ones(nn), lower=-20, upper=30, res_ref=1.0)

        bal.add_balance('eta', units='rad', val=0.01*np.ones(nn), eq_units=None, lhs_name='CM',
                        lower=-30, upper=30, res_ref=1.0)

        self.connect('aero.CL', 'alpha_eta_balance.CL')
        self.connect('aero.CD', 'thrust_eq_comp.CD')
        self.connect('aero.CM', 'alpha_eta_balance.CM')
        self.connect('CL_eq', ('alpha_eta_balance.CL_eq'))

        self.linear_solver = om.DirectSolver()
        self.nonlinear_solver = om.NewtonSolver()
        self.nonlinear_solver.options['atol'] = 1e-14
        self.nonlinear_solver.options['rtol'] = 1e-14
        self.nonlinear_solver.options['solve_subsystems'] = True
        self.nonlinear_solver.options['err_on_non_converge'] = True
        self.nonlinear_solver.options['max_sub_solves'] = 10
        self.nonlinear_solver.options['maxiter'] = 150
        self.nonlinear_solver.options['iprint'] = -1
        self.nonlinear_solver.linesearch = om.BoundsEnforceLS()
        self.nonlinear_solver.linesearch.options['print_bound_enforce'] = True



import unittest

try:
    import matplotlib.pyplot as plt
    plt.switch_backend('Agg')
except ImportError:
    plt = None

from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse


@use_tempdirs
class TestTwoPhaseCannonballODEOutputLinkage(unittest.TestCase):

    @require_pyoptsparse(optimizer='SLSQP')
    def test_link_fixed_states_final_to_initial(self):
        """ Test that linking phases with states that are fixed at the linkage point raises an exception. """

        import openmdao.api as om

        import dymos as dm
        from dymos.examples.cannonball.size_comp import CannonballSizeComp
        from dymos.examples.cannonball.cannonball_ode import CannonballODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=False,  units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2')
        ascent.add_parameter('mass', targets=['m'], units='kg')

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=False, fix_final=True,  units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=True, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2')
        descent.add_parameter('mass', targets=['m'], units='kg')

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'h', 'gam'], connected=False)

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        # Issue Connections
        p.model.connect('external_params.radius', 'size_comp.radius')
        p.model.connect('external_params.dens', 'size_comp.dens')

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link final '
                                           'value of "gam" in ascent to initial value of "gam" in '
                                           'descent. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_link_fixed_states_final_to_final(self):
        """ Test that linking phases with states that are fixed at the linkage point raises an exception. """

        import openmdao.api as om

        import dymos as dm
        from dymos.examples.cannonball.size_comp import CannonballSizeComp
        from dymos.examples.cannonball.cannonball_ode import CannonballODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=True, units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2')
        ascent.add_parameter('mass', targets=['m'], units='kg')

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=True, fix_final=True, units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=False, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2')
        descent.add_parameter('mass', targets=['m'], units='kg')

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'gam'], connected=False)

        traj.add_linkage_constraint(phase_a='ascent', phase_b='descent', var_a='h', var_b='h',
                                    loc_a='final', loc_b='final')

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        # Issue Connections
        p.model.connect('external_params.radius', 'size_comp.radius')
        p.model.connect('external_params.dens', 'size_comp.dens')

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link final '
                                           'value of "h" in ascent to final value of "h" in '
                                           'descent. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_link_fixed_states_initial_to_initial(self):
        """ Test that linking phases with states that are fixed at the linkage point raises an exception. """

        import openmdao.api as om

        import dymos as dm
        from dymos.examples.cannonball.size_comp import CannonballSizeComp
        from dymos.examples.cannonball.cannonball_ode import CannonballODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, duration_bounds=(1, 100), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=True, units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2')
        ascent.add_parameter('mass', targets=['m'], units='kg')

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(.5, 100), duration_bounds=(.5, 100),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=True, fix_final=True, units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=False, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2')
        descent.add_parameter('mass', targets=['m'], units='kg')

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'gam'], connected=False)

        traj.add_linkage_constraint(phase_a='ascent', phase_b='descent', var_a='h', var_b='h',
                                    loc_a='initial', loc_b='initial')

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        # Issue Connections
        p.model.connect('external_params.radius', 'size_comp.radius')
        p.model.connect('external_params.dens', 'size_comp.dens')

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link initial '
                                           'value of "h" in ascent to initial value of "h" in '
                                           'descent. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_link_fixed_times_final_to_initial(self):
        """ Test that linking phases with times that are fixed at the linkage point raises an exception. """

        import openmdao.api as om

        import dymos as dm
        from dymos.examples.cannonball.size_comp import CannonballSizeComp
        from dymos.examples.cannonball.cannonball_ode import CannonballODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, fix_duration=True, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=False, units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2')
        ascent.add_parameter('mass', targets=['m'], units='kg')

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(fix_initial=True, fix_duration=True, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=False, fix_final=True, units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=False, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, units='m/s', rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2')
        descent.add_parameter('mass', targets=['m'], units='kg')

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'h', 'gam'], connected=False)

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        # Issue Connections
        p.model.connect('external_params.radius', 'size_comp.radius')
        p.model.connect('external_params.dens', 'size_comp.dens')

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link final '
                                           'value of "time" in ascent to initial value of "time" in '
                                           'descent. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')

    @require_pyoptsparse(optimizer='SLSQP')
    def test_link_bounded_times_final_to_initial(self):
        """ Test that linking phases with times that are fixed at the linkage point raises an exception. """
        import openmdao.api as om

        import dymos as dm
        from dymos.examples.cannonball.size_comp import CannonballSizeComp
        from dymos.examples.cannonball.cannonball_ode import CannonballODE

        p = om.Problem(model=om.Group())

        p.driver = om.pyOptSparseDriver()
        p.driver.options['optimizer'] = 'SLSQP'
        p.driver.declare_coloring()

        p.model.add_subsystem('size_comp', CannonballSizeComp(), promotes_inputs=['radius', 'dens'])
        p.model.set_input_defaults('dens', val=7.87, units='g/cm**3')
        p.model.add_design_var('radius', lower=0.01, upper=0.10, ref0=0.01, ref=0.10, units='m')

        traj = p.model.add_subsystem('traj', dm.Trajectory())

        transcription = dm.Radau(num_segments=5, order=3, compressed=True)
        ascent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        ascent = traj.add_phase('ascent', ascent)

        # All initial states except flight path angle are fixed
        # Final flight path angle is fixed (we will set it to zero so that the phase ends at apogee)
        ascent.set_time_options(fix_initial=True, duration_bounds=(10, 10), duration_ref=100, units='s')
        ascent.add_state('r', fix_initial=True, fix_final=False, units='m', rate_source='r_dot')
        ascent.add_state('h', fix_initial=True, fix_final=False, units='m', rate_source='h_dot')
        ascent.add_state('gam', fix_initial=False, fix_final=True, units='rad', rate_source='gam_dot')
        ascent.add_state('v', fix_initial=False, fix_final=False, rate_source='v_dot')

        ascent.add_parameter('S', targets=['S'], units='m**2')
        ascent.add_parameter('mass', targets=['m'], units='kg')

        # Limit the muzzle energy
        ascent.add_boundary_constraint('ke', loc='initial', units='J',
                                       upper=400000, lower=0, ref=100000, shape=(1,))

        # Second Phase (descent)
        transcription = dm.GaussLobatto(num_segments=5, order=3, compressed=True)
        descent = dm.Phase(ode_class=CannonballODE, transcription=transcription)

        traj.add_phase('descent', descent)

        # All initial states and time are free (they will be linked to the final states of ascent.
        # Final altitude is fixed (we will set it to zero so that the phase ends at ground impact)
        descent.set_time_options(initial_bounds=(10, 10), duration_bounds=(10, 10),
                                 duration_ref=100, units='s')
        descent.add_state('r', units='m', rate_source='r_dot')
        descent.add_state('h', fix_initial=False, fix_final=True, units='m', rate_source='h_dot')
        descent.add_state('gam', fix_initial=False, fix_final=False, units='rad', rate_source='gam_dot')
        descent.add_state('v', fix_initial=False, fix_final=False, rate_source='v_dot')

        descent.add_parameter('S', targets=['S'], units='m**2')
        descent.add_parameter('mass', targets=['m'], units='kg')

        descent.add_objective('r', loc='final', scaler=-1.0)

        # Add internally-managed design parameters to the trajectory.
        traj.add_parameter('CD',
                           targets={'ascent': ['CD'], 'descent': ['CD']},
                           val=0.5, units=None, opt=False)

        # Add externally-provided design parameters to the trajectory.
        # In this case, we connect 'm' to pre-existing input parameters named 'mass' in each phase.
        traj.add_parameter('m', units='kg', val=1.0,
                           targets={'ascent': 'mass', 'descent': 'mass'})

        # In this case, by omitting targets, we're connecting these parameters to parameters
        # with the same name in each phase.
        traj.add_parameter('S', units='m**2', val=0.005)

        # Link Phases (link time and all state variables)
        # Note velocity is not included here.  Doing so is equivalent to linking kinetic energy,
        # and causes a duplicate row in the constraint jacobian.
        traj.link_phases(phases=['ascent', 'descent'], vars=['time', 'r', 'h', 'gam'], connected=False)

        traj.add_linkage_constraint('ascent', 'descent', 'ke', 'ke',
                                    ref=100000, connected=False)

        # Issue Connections
        p.model.connect('external_params.radius', 'size_comp.radius')
        p.model.connect('external_params.dens', 'size_comp.dens')

        p.model.connect('size_comp.mass', 'traj.parameters:m')
        p.model.connect('size_comp.S', 'traj.parameters:S')

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link final '
                                           'value of "time" in ascent to initial value of "time" in '
                                           'descent. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')

    def test_link_parameters_fixed_to_fixed(self):
        """ Linking non-opt parameters across phases by connection should work.
         But linking them via constraint should not (no freedom to move them to satisfy the constraint).
         """
        import openmdao.api as om
        import dymos as dm
        from dymos.examples.finite_burn_orbit_raise.finite_burn_eom import FiniteBurnODE

        p = om.Problem()

        traj = dm.Trajectory()
        p.model.add_subsystem('traj', subsys=traj)

        # First Phase (burn)
        burn1 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=4, order=3))

        traj.add_phase('burn1', burn1)

        burn1.set_time_options(fix_initial=True, duration_bounds=(.5, 10), units='TU')
        burn1.add_state('r', fix_initial=True, fix_final=False,
                        rate_source='r_dot', units='DU')
        burn1.add_state('theta', fix_initial=True, fix_final=False,
                        rate_source='theta_dot', units='rad')
        burn1.add_state('vr', fix_initial=True, fix_final=False,
                        rate_source='vr_dot', units='DU/TU')
        burn1.add_state('vt', fix_initial=True, fix_final=False,
                        rate_source='vt_dot', units='DU/TU')
        burn1.add_state('accel', fix_initial=True, fix_final=False,
                        rate_source='at_dot', units='DU/TU**2')
        burn1.add_state('deltav', fix_initial=True, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn1.add_control('u1',  rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn1.add_parameter('c', opt=False, val=1.5, units='DU/TU')

        # Second Phase (Coast)

        coast = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=10, order=3))

        traj.add_phase('coast', coast)

        coast.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), duration_ref=10, units='TU')
        coast.add_state('r', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='r_dot', units='DU')
        coast.add_state('theta', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='theta_dot', units='rad')
        coast.add_state('vr', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vr_dot', units='DU/TU')
        coast.add_state('vt', fix_initial=False, fix_final=False, defect_scaler=100.0,
                        rate_source='vt_dot', units='DU/TU')
        coast.add_state('accel', fix_initial=True, fix_final=True, ref=1.0E-12, defect_ref=1.0E-12,
                        rate_source='at_dot', units='DU/TU**2')
        coast.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        coast.add_control('u1', opt=False, val=0.0, units='deg')
        coast.add_parameter('c', opt=False, val=1.5, units='DU/TU')

        # Third Phase (burn)

        burn2 = dm.Phase(ode_class=FiniteBurnODE, transcription=dm.GaussLobatto(num_segments=3, order=3))

        traj.add_phase('burn2', burn2)

        burn2.set_time_options(initial_bounds=(0.5, 20), duration_bounds=(.5, 10), initial_ref=10, units='TU')
        burn2.add_state('r', fix_initial=False, fix_final=True,
                        rate_source='r_dot', units='DU')
        burn2.add_state('theta', fix_initial=False, fix_final=False,
                        rate_source='theta_dot', units='rad')
        burn2.add_state('vr', fix_initial=False, fix_final=True,
                        rate_source='vr_dot', units='DU/TU')
        burn2.add_state('vt', fix_initial=False, fix_final=True,
                        rate_source='vt_dot', units='DU/TU')
        burn2.add_state('accel', fix_initial=False, fix_final=False, defect_ref=1.0E-6,
                        rate_source='at_dot', units='DU/TU**2')
        burn2.add_state('deltav', fix_initial=False, fix_final=False,
                        rate_source='deltav_dot', units='DU/TU')
        burn2.add_control('u1', rate_continuity=True, rate2_continuity=True,
                          units='deg', scaler=0.01, lower=-30, upper=30)
        burn2.add_parameter('c', opt=False, val=1.5, units='DU/TU')

        burn2.add_objective('deltav', loc='final')

        # Link Phases
        traj.link_phases(phases=['burn1', 'coast', 'burn2'],
                         vars=['time', 'r', 'theta', 'vr', 'vt', 'deltav', 'c'])

        with self.assertRaises(ValueError) as e:
            p.setup()

        self.assertEqual(str(e.exception), 'Invalid linkage in Trajectory traj: Cannot link final value of "c" in '
                                           'burn1 to initial value of "c" in coast. Values on both sides of the '
                                           'linkage are fixed and the linkage is enforced via constraint. Either link '
                                           'the variables via connection or make the variables design variables on at '
                                           'least one side of the connection.')


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import openmdao.api as om
import numpy as np


class LaunchVehicleODE(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int,
                             desc='Number of nodes to be evaluated in the RHS')

        self.options.declare('g', types=float, default=9.80665,
                             desc='Gravitational acceleration, m/s**2')

        self.options.declare('rho_ref', types=float, default=1.225,
                             desc='Reference atmospheric density, kg/m**3')

        self.options.declare('h_scale', types=float, default=8.44E3,
                             desc='Reference altitude, m')

        self.options.declare('CD', types=float, default=0.5,
                             desc='coefficient of drag')

        self.options.declare('S', types=float, default=7.069,
                             desc='aerodynamic reference area (m**2)')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('y',
                       val=np.zeros(nn),
                       desc='altitude',
                       units='m')

        self.add_input('vx',
                       val=np.zeros(nn),
                       desc='x velocity',
                       units='m/s')

        self.add_input('vy',
                       val=np.zeros(nn),
                       desc='y velocity',
                       units='m/s')

        self.add_input('m',
                       val=np.zeros(nn),
                       desc='mass',
                       units='kg')

        self.add_input('theta',
                       val=np.zeros(nn),
                       desc='pitch angle',
                       units='rad')

        self.add_input('thrust',
                       val=2100000 * np.ones(nn),
                       desc='thrust',
                       units='N')

        self.add_input('Isp',
                       val=265.2 * np.ones(nn),
                       desc='specific impulse',
                       units='s')
        # Outputs
        self.add_output('xdot',
                        val=np.zeros(nn),
                        desc='velocity component in x',
                        units='m/s')

        self.add_output('ydot',
                        val=np.zeros(nn),
                        desc='velocity component in y',
                        units='m/s')

        self.add_output('vxdot',
                        val=np.zeros(nn),
                        desc='x acceleration magnitude',
                        units='m/s**2')

        self.add_output('vydot',
                        val=np.zeros(nn),
                        desc='y acceleration magnitude',
                        units='m/s**2')

        self.add_output('mdot',
                        val=np.zeros(nn),
                        desc='mass rate of change',
                        units='kg/s')

        self.add_output('rho',
                        val=np.zeros(nn),
                        desc='density',
                        units='kg/m**3')

        # Setup partials
        # Complex-step derivatives
        self.declare_coloring(wrt='*', method='cs', show_sparsity=False)

    def compute(self, inputs, outputs):

        theta = inputs['theta']
        cos_theta = np.cos(theta)
        sin_theta = np.sin(theta)
        vx = inputs['vx']
        vy = inputs['vy']
        m = inputs['m']
        F_T = inputs['thrust']
        Isp = inputs['Isp']
        y = inputs['y']

        g = self.options['g']
        rho_ref = self.options['rho_ref']
        h_scale = self.options['h_scale']

        CDA = self.options['CD'] * self.options['S']

        outputs['rho'] = rho_ref * np.exp(-y / h_scale)
        outputs['xdot'] = vx
        outputs['ydot'] = vy
        outputs['vxdot'] = (F_T * cos_theta - 0.5 * CDA * outputs['rho'] * vx**2) / m
        outputs['vydot'] = (F_T * sin_theta - 0.5 * CDA * outputs['rho'] * vy**2) / m - g
        outputs['mdot'] = -F_T / (g * Isp)



import numpy as np

_lgl_cache = {}
""" Cache for the LGL nodes and weights, keyed by n. """


def _lgl(n, tol=np.finfo(float).eps):
    """
    Returns the Legendre-Gauss-Lobatto nodes and weights for a Jacobi Polynomial with n abscissae.

    The nodes are on the range [-1, 1].

    Based on the routine written by Greg von Winckel (License follows)

    Copyright (c) 2009, Greg von Winckel
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

      * Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.
      * Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in
        the documentation and/or other materials provided with the distribution

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR aux_outputs
    PARTICULAR PURPOSE  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
    OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
    OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

    Parameters
    ----------
    n : int
        The number of LGL nodes requested.  The order of the polynomial is n-1.
    tol : float
        The tolerance to which the location of the nodes should be converged.

    Returns
    -------
    x : numpy.array
        An array of the LGL nodes for a polynomial of the given order.

    w : numpy.array
        An array of the corresponding LGL weights at the nodes in x.
    """
    n = n - 1
    n1 = n + 1
    n2 = n + 2
    # Get the initial guesses from the Chebyshev nodes
    x = np.cos(np.pi * (2 * np.arange(1, n2) - 1) / (2 * n1))
    P = np.zeros([n1, n1])
    # Compute P_(n) using the recursion relation
    # Compute its first and second derivatives and
    # update x using the Newton-Raphson method.
    xold = 2

    for i in range(100):
        if np.all(np.abs(x - xold) <= tol):
            break
        xold = x
        P[:, 0] = 1.0
        P[:, 1] = x

        for k in range(2, n1):
            P[:, k] = ((2 * k - 1) * x * P[:, k - 1] - (k - 1) * P[:, k - 2]) / k

        x = xold - (x * P[:, n] - P[:, n - 1]) / (n1 * P[:, n])
    else:
        raise RuntimeError(f'Failed to converge LGL nodes for order {n}')

    x.sort()

    w = 2 / (n * n1 * P[:, n]**2)

    return x, w


def lgl(n):
    """
    Retrieve the lgl nodes and weights for n nodes.

    Results are cached to avoid repeated calculation of nodes and weights for a given n.

    Parameters
    ----------
    n : int
        Node number.

    Returns
    -------
    float
        Tuple with lgl nodes and weights.
    """
    if n not in _lgl_cache:
        _lgl_cache[n] = _lgl(n)
    return _lgl_cache[n]



import numpy as np

import openmdao.api as om
from dymos.examples.min_time_climb.aero.dynamic_pressure_comp import DynamicPressureComp
from dymos.examples.min_time_climb.aero.lift_drag_force_comp import LiftDragForceComp
from dymos.models.atmosphere import USatm1976Comp
from dymos.models.eom import FlightPathEOM2D
from .water_engine_comp import WaterEngine


class WaterPropulsionODE(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=int)
        self.options.declare('ballistic', types=bool, default=False,
                             desc='If True, neglect propulsion system.')

    def setup(self):
        nn = self.options['num_nodes']

        self.add_subsystem(name='atmos',
                           subsys=USatm1976Comp(num_nodes=nn))

        if not self.options['ballistic']:
            self.add_subsystem(name='water_engine',
                               subsys=WaterEngine(num_nodes=nn),
                               promotes_inputs=['rho_w'])

            self.add_subsystem(name='mass_adder',
                               subsys=_MassAdder(num_nodes=nn),
                               promotes_inputs=['rho_w'])

        self.add_subsystem(name='dynamic_pressure',
                           subsys=DynamicPressureComp(num_nodes=nn))

        self.add_subsystem(name='aero',
                           subsys=LiftDragForceComp(num_nodes=nn))

        self.add_subsystem(name='eom',
                           subsys=FlightPathEOM2D(num_nodes=nn))

        self.connect('atmos.rho', 'dynamic_pressure.rho')
        self.connect('dynamic_pressure.q', 'aero.q')

        self.connect('aero.f_drag', 'eom.D')
        self.connect('aero.f_lift', 'eom.L')

        if not self.options['ballistic']:
            self.connect('atmos.pres', 'water_engine.p_a')
            self.connect('water_engine.F', 'eom.T')
            self.connect('mass_adder.m', 'eom.m')


class _MassAdder(om.ExplicitComponent):

    def initialize(self):
        self.options.declare('num_nodes', types=int)

    def setup(self):
        nn = self.options['num_nodes']

        self.add_input('m_empty', val=np.zeros(nn), desc='empty mass', units='kg')
        self.add_input('V_w', val=1e-3*np.ones(nn), desc='water volume', units='m**3')
        self.add_input('rho_w', val=1e3*np.ones(nn), desc="water density", units='kg/m**3')

        self.add_output('m', val=np.zeros(nn), desc='total mass', units='kg')

        ar = np.arange(nn)
        self.declare_partials('*', '*', cols=ar, rows=ar)

    def compute(self, inputs, outputs):
        outputs['m'] = inputs['m_empty'] + inputs['rho_w']*inputs['V_w']

    def compute_partials(self, inputs, jacobian):
        jacobian['m', 'm_empty'] = 1
        jacobian['m', 'rho_w'] = inputs['V_w']
        jacobian['m', 'V_w'] = inputs['rho_w']


if __name__ == '__main__':
    p = om.Problem()
    p.model = WaterPropulsionODE(num_nodes=1)
    p.setup()
    p.check_config(checks=['unconnected_inputs'], out_file=None)
    p.final_setup()



import functools
from typing import Union
import numpy as np
from numpy.typing import ArrayLike
import openmdao.api as om

try:
    from numba import njit
except ImportError:
    # If numba is not available, just write a dummy njit wrapper.
    # Code will still run at a significant performance hit.
    def njit(*args, **kwargs):
        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                # Call the original function
                return func(*args, **kwargs)
            return wrapper
        return decorator


from ...utils.lgl import lgl
from ...utils.misc import get_rate_units


@njit()
def _compute_dl_dg(tau: float,
                   taus: ArrayLike,
                   l: ArrayLike,  # noqa: E741, allow ambiguous variable name 'l'
                   dl_dg: ArrayLike,
                   d2l_dg2: ArrayLike,
                   d3l_dg3: Union[ArrayLike, None] = None,):
    """Compute the Lagrange polynomials and the first three derivatives wrt g at the nodes.

    This function achieves good performance with numba.njit (and uses a do-nothing njit decorator
    if numba is not available).

    Parameters
    ----------
    tau : float
        The value of the independent variable at which the interpolation is being requested.
    taus : ArrayLike
        An n-vector giving location of the polynomial nodes in the independent variable dimension.
    l : ArrayLike
        An n-vector giving the value of the Lagrange polynomial at each node.
    dl_dg : ArrayLike
        An n x n - vector giving the derivative of l wrt g.
    d2l_dg2 : ArrayLike
        An n x n x n - vector giving the second derivative of l wrt g.
    d3l_dg3 : ArrayLike
        An n x n x n - vector giving the second derivative of l wrt g.
    """
    n = len(taus)
    g = tau - taus

    for i in range(n):
        i_save = g[i]
        g[i] = 1.0
        l[i] = np.prod(g)
        # dl_dg is symmetric, so fill two elements at once
        if dl_dg is not None:
            for j in range(i+1, n):
                j_save = g[j]
                g[j] = 1.0
                val = np.prod(g)
                dl_dg[i, j] = dl_dg[j, i] = val
                if d2l_dg2 is not None:
                    for k in range(n):
                        if k != i and k != j:
                            k_save = g[k]
                            g[k] = 1.0
                            d2l_dg2[i, j, k] = d2l_dg2[j, i, k] = np.prod(g)
                            if d3l_dg3 is not None:
                                # We only need this if derivs of second deriv wrt tau are needed.
                                for ii in range(k + 1, n):
                                    if ii not in {i, j, k}:
                                        ii_save = g[ii]
                                        g[ii] = 1
                                        d3l_dg3[i, j, k, ii] = d3l_dg3[j, i, k, ii] = \
                                            d3l_dg3[i, j, ii, k] = d3l_dg3[j, i, ii, k] = np.prod(g)
                                        g[ii] = ii_save
                            g[k] = k_save
                g[j] = j_save
        g[i] = i_save


class BarycentricControlInterpComp(om.ExplicitComponent):
    """
    A component which interpolates control values in 1D using Vandermonde interpolation.

    Takes training values for control variables at given _input_ nodes,
    broadcasts them to _discretization_ nodes, and then interpolates the discretization values
    to provide a control variable at a given segment tau or phase tau.

    For dynamic controls, the current segment is given as a discrete input and the interpolation is
    a smooth polynomial along the given segment.

    OpenMDAO assumes sizes of variables at setup time, and we don't want to need to change the
    size of the control input nodes when we evaluate different segments. Instead, this component
    will take in the control values of all segments and internally use the appropriate one.

    Parameters
    ----------
    grid_data : GridData
        A GridData instance that details information on how the control input and discretization
        nodes are layed out.
    control_options : dict of {str: ControlOptionsDictionary}
        A mapping that maps the name of each control to a ControlOptionsDictionary of its options.
    time_units : str
        The time units pertaining to the control rates.
    standalone_mode : bool
        If True, this component runs its configuration steps during setup. This is useful for
        unittests in which the component does not exist in a larger group.
    compute_derivs : bool
        Set to True if derivatives of rate2 need to be computed, otherwise False.
    **kwargs : dict, optional
        Keyword arguments passed to ExplicitComponent.
    """
    def __init__(self, grid_data, control_options=None,
                 time_units=None, standalone_mode=False, compute_derivs=True, **kwargs):
        self._grid_data = grid_data
        self._control_options = {} if control_options is None else control_options
        self._time_units = time_units
        self._standalone_mode = standalone_mode
        self._compute_derivs = compute_derivs
        self._under_complex_step_prev = False
        self._taus_seg = {}

        self._inputs_hash_cache = None

        # Cache formatted strings: { control_name : (input_name, output_name) }
        self._control_io_names = {}

        # The Lagrange interpolation matrix L_id maps control values as given at the input nodes
        # to values at the discretization nodes.
        num_disc_nodes = grid_data.subset_num_nodes['control_disc']
        num_input_nodes = grid_data.subset_num_nodes['control_input']
        self._L_id = {}
        self._L_id['controls'] = np.zeros((num_disc_nodes, num_input_nodes), dtype=float)
        self._L_id['controls'][np.arange(num_disc_nodes, dtype=int),
                               self._grid_data.input_maps['dynamic_control_input_to_disc']] = 1.0

        super().__init__(**kwargs)

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('segment_index', default=None, types=int, desc='index of the current segment')

    def set_segment_index(self, idx, alloc_complex=False):
        """
        Set the index of the segment being interpolated.

        Parameters
        ----------
        idx : int
            The index of the segment being interpolated.
        alloc_complex : bool
            If True, allocate storage for complex step.
        """
        self.options['segment_index'] = idx

        i1, i2 = self._grid_data.subset_segment_indices['control_disc'][idx, :]
        indices = self._grid_data.subset_node_indices['control_disc'][i1:i2]
        taus_seg = self._grid_data.node_stau[indices]

        n = len(taus_seg)
        ptaus = {}
        for control_name, options in self._control_options.items():
            if options['control_type'] == 'polynomial':
                order = options['order']
                ptaus[control_name] = lgl(order + 1)[0]
        self._compute_barycentric_weights(taus_seg, ptaus)
        dtype = complex if alloc_complex else float

        # The arrays pertaining to the collocated controls are stored with the 'controls' key
        self._l = {'controls': np.ones(n, dtype=dtype)}
        self._dl_dg = {'controls': np.zeros((n, n), dtype=dtype)}
        self._d2l_dg2 = {'controls': np.zeros((n, n, n), dtype=dtype)}
        self._dl_dtau = {'controls': np.ones((n, 1), dtype=dtype)}
        self._d2l_dtau2 = {'controls': np.ones((n, 1), dtype=dtype)}

        if self._compute_derivs:
            self._d3l_dg3 = {'controls': np.zeros((n, n, n, n), dtype=dtype)}
            self._d3l_dtau3 = {'controls': np.ones((n, 1), dtype=dtype)}
        else:
            self._d3l_dg3 = {'controls': None}
            self._d3l_dtau3 = {'controls': None}

        self._taus_seg['controls'] = taus_seg

        # Arrays pertaining to polynomial controls are stored with their name as a key
        for pc_name in ptaus:
            pc_options = self._control_options[pc_name]
            n = pc_options['order'] + 1
            self._taus_seg[pc_name] = lgl(n)[0]
            self._l[pc_name] = np.ones(n, dtype=dtype)
            self._dl_dg[pc_name] = np.zeros((n, n), dtype=dtype)
            self._d2l_dg2[pc_name] = np.zeros((n, n, n), dtype=dtype)
            self._dl_dtau[pc_name] = np.ones((n, 1), dtype=dtype)
            self._d2l_dtau2[pc_name] = np.ones((n, 1), dtype=dtype)

            if self._compute_derivs:
                self._d3l_dg3[pc_name] = np.zeros((n, n, n, n), dtype=dtype)
                self._d3l_dtau3[pc_name] = np.ones((n, 1), dtype=dtype)
            else:
                self._d3l_dg3[pc_name] = None
                self._d3l_dtau3[pc_name] = None

    def _configure_controls(self):
        gd = self._grid_data

        self._disc_node_idxs_by_segment = []
        self._input_node_idxs_by_segment = []

        first_disc_node_in_seg = 0

        for seg_idx in range(gd.num_segments):
            # Number of control discretization nodes per segment
            ncdnps = gd.subset_num_nodes_per_segment['control_disc'][seg_idx]
            ar_control_disc_nodes = np.arange(ncdnps, dtype=int)
            disc_idxs_in_seg = first_disc_node_in_seg + ar_control_disc_nodes
            first_disc_node_in_seg += ncdnps

            # The indices of the discretization node u vector pertaining to the given segment
            self._disc_node_idxs_by_segment.append(disc_idxs_in_seg)

            # The indices of the input u vector pertaining to the given segment
            self._input_node_idxs_by_segment.append(gd.input_maps['dynamic_control_input_to_disc'][disc_idxs_in_seg])

        if not self._control_options:
            return

        num_uhat_nodes = gd.subset_num_nodes['control_input']
        for control_name, options in self._control_options.items():
            if options['control_type'] == 'full':
                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                uhat_shape = (num_uhat_nodes,) + shape
                output_shape = (1,) + shape
                self.add_input(input_name, shape=uhat_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)
                self.declare_partials(of=output_name, wrt=input_name)
                self.declare_partials(of=output_name, wrt='stau')
                self.declare_partials(of=rate_name, wrt=input_name)
                self.declare_partials(of=rate_name, wrt='stau')
                self.declare_partials(of=rate_name, wrt='dstau_dt')

                if self._compute_derivs:
                    self.declare_partials(of=rate2_name, wrt=input_name)
                    self.declare_partials(of=rate2_name, wrt='stau')
                    self.declare_partials(of=rate2_name, wrt='dstau_dt')

            elif options['control_type'] == 'polynomial':
                order = options['order']

                shape = options['shape']
                units = options['units']
                input_name = f'controls:{control_name}'
                output_name = f'control_values:{control_name}'
                rate_name = f'control_rates:{control_name}_rate'
                rate2_name = f'control_rates:{control_name}_rate2'
                rate_units = get_rate_units(units, self._time_units)
                rate2_units = get_rate_units(units, self._time_units, deriv=2)
                input_shape = (order + 1,) + shape
                output_shape = (1,) + shape
                self.add_input(input_name, shape=input_shape, units=units)
                self.add_output(output_name, shape=output_shape, units=units)
                self.add_output(rate_name, shape=output_shape, units=rate_units)
                self.add_output(rate2_name, shape=output_shape, units=rate2_units)
                self._control_io_names[control_name] = (input_name, output_name, rate_name, rate2_name)
                self.declare_partials(of=output_name, wrt=input_name)
                self.declare_partials(of=output_name, wrt='ptau')
                self.declare_partials(of=rate_name, wrt=input_name)
                self.declare_partials(of=rate_name, wrt='ptau')
                self.declare_partials(of=rate_name, wrt='t_duration')

                if self._compute_derivs:
                    self.declare_partials(of=rate2_name, wrt=input_name)
                    self.declare_partials(of=rate2_name, wrt='ptau')
                    self.declare_partials(of=rate2_name, wrt='t_duration')

    def setup(self):
        """
        Perform the I/O creation if operating in _standalone_mode.
        """
        if self._standalone_mode:
            self.configure_io()

    def configure_io(self):
        """
        I/O creation is delayed until configure so we can determine shape and units for the controls.
        """

        # self.add_discrete_input('segment_index', val=0, desc='index of the segment')
        self.add_input('stau', shape=(1,), units=None)
        self.add_input('dstau_dt', val=1.0, units=f'1/{self._time_units}')
        self.add_input('t_duration', val=1.0, units=self._time_units)
        self.add_input('ptau', shape=(1,), units=None)

        self._configure_controls()

    def _compute_barycentric_weights(self, taus: ArrayLike, ptaus: dict):
        """Computes the barycentric weights given a set of nodes.

        Parameters
        ----------
        taus : ArrayLike
            The nodes of the collocated control polynomial.
        ptaus : dict[ArrayLike]
            The nodes of each polynomial control polynomial.
        """
        n = len(taus)
        self._w_b = {'controls': np.ones((n, 1))}

        for j in range(n):
            self._w_b['controls'][j, 0] = 1. / np.prod(taus[j] - np.delete(taus, j))

        for pc_name, _ptaus in ptaus.items():
            n = len(_ptaus)
            self._w_b[pc_name] = np.ones((n, 1))
            for j in range(n):
                self._w_b[pc_name][j, 0] = 1. / np.prod(_ptaus[j] - np.delete(_ptaus, j))

    def _compute_controls(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute interpolated control values and rates for the collocated controls.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete_inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete_outputs.
        """
        seg_idx = self.options['segment_index']
        stau = inputs['stau']
        dstau_dt = inputs['dstau_dt']

        disc_node_idxs = self._disc_node_idxs_by_segment[seg_idx]
        input_node_idxs = self._input_node_idxs_by_segment[seg_idx]
        # V_stau = np.vander(stau, N=n, increasing=True)
        # dV_stau, dV2_stau, _ = self._dvander(V_stau)
        taus_seg = self._taus_seg['controls']

        # Retrieve the storage vectors that pertain to the collocated controls
        l = self._l['controls']  # noqa: E741, allow ambiguous variable name 'l'
        dl_dg = self._dl_dg['controls']
        d2l_dg2 = self._d2l_dg2['controls']

        dl_dstau = self._dl_dtau['controls']
        d2l_dstau2 = self._d2l_dtau2['controls']
        w_b = self._w_b['controls']

        # _compute_dl_dg(stau, taus_seg, l, dl_dg, d2l_dg2, d3l_dg3=None)
        _compute_dl_dg(tau=stau, taus=taus_seg, l=l, dl_dg=dl_dg, d2l_dg2=d2l_dg2, d3l_dg3=None)

        # Equivalent of multiplying dl_dg @ dg_dtau, where dg_dtau is a column vector of n ones.
        dl_dstau[...] = np.sum(dl_dg, axis=-1, keepdims=True)

        # d2l_dg @ dg_dtau + dl_dg @ d2g_dtau2 but d2g_dtau2 is zero.
        d2l_dstau2[...] = np.sum(np.sum(d2l_dg2, axis=-1), axis=-1, keepdims=True)

        L_id = self._L_id['controls']
        L_seg = L_id[disc_node_idxs[0]:disc_node_idxs[0] + len(disc_node_idxs),
                     input_node_idxs[0]:input_node_idxs[0] + len(input_node_idxs)]

        seg_idx = self.options['segment_index']
        ptau = inputs['ptau']
        dptau_dt = 2. / inputs['t_duration']

        disc_node_idxs = self._disc_node_idxs_by_segment[seg_idx]
        input_node_idxs = self._input_node_idxs_by_segment[seg_idx]

        for control_name, options in self._control_options.items():
            if options['control_type'] == 'full':
                input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]

                # Translate the input nodes to the discretization nodes.
                u_hat = np.dot(L_seg, inputs[input_name][input_node_idxs])

                # Perform a row_wise multiplication of w_b and u_hat
                wbuhat = np.einsum("ij,i...->i...", w_b, u_hat)

                outputs[output_name] = np.einsum('i...,i...->...', l, wbuhat)
                outputs[rate_name] = wbuhat.T @ dl_dstau * dstau_dt
                outputs[rate2_name] = wbuhat.T @ d2l_dstau2 * dstau_dt ** 2

            else:
                # Retrieve the storage vectors that pertain to the polynomial control
                taus_seg = self._taus_seg[control_name]
                l = self._l[control_name]  # noqa: E741, allow ambiguous variable name 'l'
                dl_dg = self._dl_dg[control_name]
                d2l_dg2 = self._d2l_dg2[control_name]

                dl_dptau = self._dl_dtau[control_name]
                d2l_dptau2 = self._d2l_dtau2[control_name]
                w_b = self._w_b[control_name]

                _compute_dl_dg(ptau, taus_seg, l, dl_dg, d2l_dg2)

                # Equivalent of multiplying dl_dg @ dg_dtau, where dg_dtau is a column vector of n ones.
                dl_dptau[...] = np.sum(dl_dg, axis=-1, keepdims=True)

                # d2l_dg @ dg_dtau + dl_dg @ d2g_dtau2 but d2g_dtau2 is zero.
                d2l_dptau2[...] = np.sum(np.sum(d2l_dg2, axis=-1), axis=-1, keepdims=True)

                input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]

                # Translate the input nodes to the discretization nodes.
                u_hat = inputs[input_name]

                # Perform a row_wise multiplication of w_b and u_hat
                wbuhat = np.einsum("ij,i...->i...", w_b, u_hat)

                # outputs[output_name] = np.einsum('i...,i...->...', l, wbuhat)
                outputs[output_name] = wbuhat.T @ l
                outputs[rate_name] = wbuhat.T @ dl_dptau * dptau_dt
                outputs[rate2_name] = wbuhat.T @ d2l_dptau2 * dptau_dt ** 2

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute interpolated control values and rates.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete_inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete_outputs.
        """
        if self.under_complex_step != self._under_complex_step_prev:
            self.set_segment_index(self.options['segment_index'], alloc_complex=self.under_complex_step)
            self._under_complex_step_prev = self.under_complex_step

        inputs_hash = inputs.get_hash()
        if inputs_hash != self._inputs_hash_cache:
            # Do the compute if our inputs have changed
            if self._control_options:
                self._compute_controls(inputs, outputs, discrete_inputs, discrete_outputs)
            self._inputs_hash_cache = inputs_hash

    def _compute_partials_controls(self, inputs, partials, discrete_inputs=None):
        """
        Compute partials of interpolated control values and rates for the collocated controls.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete_inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete_outputs.
        """
        if self._control_options:
            gd = self._grid_data
            seg_idx = self.options['segment_index']
            stau = inputs['stau']
            dstau_dt = inputs['dstau_dt']

            disc_node_idxs = self._disc_node_idxs_by_segment[seg_idx]
            input_node_idxs = self._input_node_idxs_by_segment[seg_idx]
            # V_stau = np.vander(stau, N=n, increasing=True)
            # dV_stau, dV2_stau, _ = self._dvander(V_stau)
            taus_seg = gd.node_stau[disc_node_idxs]

            # Retrieve the storage vectors that pertain to the collocated controls
            l = self._l['controls']  # noqa: E741, allow ambiguous variable name 'l'
            dl_dg = self._dl_dg['controls']
            d2l_dg2 = self._d2l_dg2['controls']
            d3l_dg3 = self._d3l_dg3['controls']
            dl_dstau = self._dl_dtau['controls']
            d2l_dstau2 = self._d2l_dtau2['controls']
            d3l_dstau3 = self._d3l_dtau3['controls']
            w_b = self._w_b['controls']

            # Update self._l, self._dl_dg, self._d2l_dg2, and self._d3l_dg3 in place.
            _compute_dl_dg(stau, taus_seg, l, dl_dg, d2l_dg2, d3l_dg3)

            # Equivalent of multiplying dl_dg @ dg_dtau, where dg_dtau is a column vector of n ones.
            dl_dstau[...] = np.sum(dl_dg, axis=-1, keepdims=True)
            d2l_dstau2[...] = np.sum(np.sum(d2l_dg2, axis=-1), axis=-1, keepdims=True)

            if self._compute_derivs:
                d3l_dstau3[...] = np.sum(np.sum(np.sum(d3l_dg3, axis=-1), axis=-1), axis=-1, keepdims=True)

            L_id = self._L_id['controls']
            L_seg = L_id[disc_node_idxs[0]:disc_node_idxs[0] + len(disc_node_idxs),
                         input_node_idxs[0]:input_node_idxs[0] + len(input_node_idxs)]

            ptau = inputs['ptau']
            t_duration = inputs['t_duration']
            dptau_dt = 2.0 / t_duration

            for control_name, options in enumerate(self._control_options):
                if options['control_type'] == 'full':
                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]

                    # Translate the input nodes to the discretization nodes.
                    u_hat = np.dot(L_seg, inputs[input_name][input_node_idxs])

                    # Perform a row_wise multiplication of w_b and u_hat
                    wbuhat = np.einsum("ij,i...->i...", w_b, u_hat)

                    # outputs[output_name] = np.einsum('i...,i...->...', l, wbuhat)
                    partials[output_name, 'stau'] = dl_dstau.T @ wbuhat

                    partials[rate_name, 'stau'] = d2l_dstau2.T @ wbuhat * dstau_dt
                    partials[rate_name, 'dstau_dt'] = partials[output_name, 'stau']

                    if self._compute_derivs:
                        partials[rate2_name, 'stau'] = d3l_dstau3.T @ wbuhat * dstau_dt ** 2
                        partials[rate2_name, 'dstau_dt'] = 2 * wbuhat.T @ d2l_dstau2 * dstau_dt

                    # Assign only thos jacobian columns due to the current segment, since
                    # other segments cannot impact interpolation in this one.
                    partials[output_name, input_name] = 0.0
                    partials[output_name, input_name][..., input_node_idxs] = \
                        ((l * w_b.T) @ L_seg)

                    partials[rate_name, input_name] = 0.0
                    partials[rate_name, input_name][..., input_node_idxs] = \
                        (dl_dstau * w_b * dstau_dt).T @ L_seg

                    if self._compute_derivs:
                        partials[rate2_name, input_name] = 0.0
                        partials[rate2_name, input_name][..., input_node_idxs] = \
                            (d2l_dstau2 * w_b * dstau_dt ** 2).T @ L_seg

                else:
                    gd = self._grid_data

                    taus_seg = self._taus_seg[control_name]

                    # Retrieve the storage vectors that pertain to the collocated controls
                    l = self._l[control_name]  # noqa: E741, allow ambiguous variable name 'l'
                    dl_dg = self._dl_dg[control_name]
                    d2l_dg2 = self._d2l_dg2[control_name]
                    d3l_dg3 = self._d3l_dg3[control_name]
                    dl_dstau = self._dl_dtau[control_name]
                    d2l_dstau2 = self._d2l_dtau2[control_name]
                    d3l_dstau3 = self._d3l_dtau3[control_name]
                    w_b = self._w_b[control_name]

                    _compute_dl_dg(ptau, taus_seg, l, dl_dg, d2l_dg2, d3l_dg3)

                    # Equivalent of multiplying dl_dg @ dg_dtau, where dg_dtau is a column vector of n ones.
                    dl_dstau[...] = np.sum(dl_dg, axis=-1, keepdims=True)
                    d2l_dstau2[...] = np.sum(np.sum(d2l_dg2, axis=-1), axis=-1, keepdims=True)

                    input_name, output_name, rate_name, rate2_name = self._control_io_names[control_name]

                    # Translate the input nodes to the discretization nodes.
                    u_hat = inputs[input_name]

                    # Perform a row_wise multiplication of w_b and u_hat
                    wbuhat = np.einsum("ij,i...->i...", w_b, u_hat)

                    # outputs[output_name] = np.einsum('i...,i...->...', l, wbuhat)

                    d_dptau_dt_d_t_duration = -2.0 / t_duration ** 2

                    partials[output_name, 'ptau'] = dl_dstau.T @ wbuhat

                    partials[rate_name, 'ptau'] = d2l_dstau2.T @ wbuhat * dptau_dt
                    partials[rate_name, 't_duration'] = partials[output_name, 'ptau'] * d_dptau_dt_d_t_duration

                    # Assign only thos jacobian columns due to the current segment, since
                    # other segments cannot impact interpolation in this one.
                    # partials[output_name, input_name] = 0.0
                    partials[output_name, input_name][...] = ((l * w_b.T).T).T

                    # partials[rate_name, input_name] = 0.0
                    partials[rate_name, input_name][...] = (dl_dstau * w_b * dptau_dt).T

                    # partials[rate2_name, input_name] = 0.0
                if self._compute_derivs:
                    d3l_dstau3[...] = np.sum(np.sum(np.sum(d3l_dg3, axis=-1), axis=-1), axis=-1, keepdims=True)
                    partials[rate2_name, 'ptau'] = d3l_dstau3.T @ wbuhat * dptau_dt ** 2
                    partials[rate2_name, 't_duration'] = 2 * wbuhat.T @ d2l_dstau2 * dptau_dt * d_dptau_dt_d_t_duration
                    partials[rate2_name, input_name][...] = (d2l_dstau2 * w_b * dptau_dt ** 2).T

    def compute_partials(self, inputs, partials, discrete_inputs=None):
        """
        Compute derivatives of interpolated control values and rates wrt the inputs.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        discrete_inputs : Vector
            Unscaled, discrete input variables keyed by variable name.
        """
        if self._control_options:
            self._compute_partials_controls(inputs, partials, discrete_inputs)



import unittest
import numpy as np
import openmdao.api as om
import dymos as dm

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs
from openmdao.utils.units import convert_units


class _TestEOM(om.Group):

    def initialize(self):
        self.options.declare('num_nodes', types=(int,))
        self.options.declare('foo_units', allow_none=True, default=None)
        self.options.declare('foo_shapes', allow_none=True, default=None)
        self.options.declare('foo_static', default=[])

    def setup(self):
        num_nodes = self.options['num_nodes']

        foo_shape = (num_nodes,) if self.options['foo_shapes'] is None \
            else (num_nodes,) + self.options['foo_shapes']['vdot_comp']

        foo_unit = 'kg' if self.options['foo_units'] is None else self.options['foo_units']['vdot_comp']

        foo_tags = ['dymos.static_target']\
            if 'vdot_comp' in self.options['foo_static'] and self.options['foo_static']['vdot_comp'] else []

        foo_shape = (1,) if 'dymos.static_target' in foo_tags else foo_shape

        vdot_comp = om.ExecComp(['vdot = g * cos(theta)',
                                 'bar = foo'],
                                vdot={'shape': (num_nodes,), 'units': 'm/s**2'},
                                g={'val': 9.80665, 'units': 'm/s**2'},
                                theta={'shape': (num_nodes,), 'units': 'rad'},
                                foo={'shape': foo_shape, 'units': foo_unit, 'tags': foo_tags},
                                bar={'shape': foo_shape, 'units': foo_unit})

        foo_shape = (num_nodes,) if self.options['foo_shapes'] is None \
            else (num_nodes,) + self.options['foo_shapes']['xdot_comp']

        foo_unit = 'kg' if self.options['foo_units'] is None else self.options['foo_units']['xdot_comp']

        foo_tags = ['dymos.static_target']\
            if 'xdot_comp' in self.options['foo_static'] and self.options['foo_static']['xdot_comp'] else []

        foo_shape = (1,) if 'dymos.static_target' in foo_tags else foo_shape

        xdot_comp = om.ExecComp(['xdot = v * sin(theta)',
                                 'bar = foo'],
                                xdot={'shape': (num_nodes,), 'units': 'm/s'},
                                v={'shape': (num_nodes,), 'units': 'm/s'},
                                theta={'shape': (num_nodes,), 'units': 'rad'},
                                foo={'shape': foo_shape, 'units': foo_unit, 'tags': foo_tags},
                                bar={'shape': foo_shape, 'units': foo_unit})

        foo_shape = (num_nodes,) if self.options['foo_shapes'] is None \
            else (num_nodes,) + self.options['foo_shapes']['ydot_comp']

        foo_unit = 'kg' if self.options['foo_units'] is None else self.options['foo_units']['ydot_comp']

        foo_tags = ['dymos.static_target']\
            if 'ydot_comp' in self.options['foo_static'] and self.options['foo_static']['ydot_comp'] else []

        foo_shape = (1,) if 'dymos.static_target' in foo_tags else foo_shape

        ydot_comp = om.ExecComp(['ydot = -v * cos(theta)',
                                 'bar = foo'],
                                ydot={'shape': (num_nodes,), 'units': 'm/s'},
                                v={'shape': (num_nodes,), 'units': 'm/s'},
                                theta={'shape': (num_nodes,), 'units': 'rad'},
                                foo={'shape': foo_shape, 'units': foo_unit, 'tags': foo_tags},
                                bar={'shape': foo_shape, 'units': foo_unit})

        self.add_subsystem('vdot_comp', vdot_comp)
        self.add_subsystem('xdot_comp', xdot_comp)
        self.add_subsystem('ydot_comp', ydot_comp)


@use_tempdirs
class TestParameterShapes(unittest.TestCase):

    def test_valid_parameters(self):
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.
        phase.set_time_val(initial=0.0, duration=2.0, units='s')
        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 9.9], units='m/s')
        phase.set_control_val('theta', [90, 90], units='deg')
        phase.set_parameter_val('foo', 5.0)

        # Run the driver to solve the problem
        p.run_driver()

        self.assertEqual((1,), phase.parameter_options['foo']['shape'])
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.parameter_vals:foo')[-1], 5.0, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.xdot_comp.foo'), 5.0*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.ydot_comp.foo'), 5.0*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.vdot_comp.foo'), 5.0*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.xdot_comp.foo'), 5.0*np.ones(20,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.ydot_comp.foo'), 5.0*np.ones(20,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.vdot_comp.foo'), 5.0*np.ones(20,), tolerance=1.0E-5)

    def test_invalid_params_different_target_shapes(self):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3),
                         ode_init_kwargs={'foo_shapes': {'xdot_comp': (2,),
                                                         'ydot_comp': (2,),
                                                         'vdot_comp': (2,)}})

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo', shape=(1,),
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        with self.assertRaises(RuntimeError) as e:
            p.setup(check=True)

        expected = ("Shape provided to parameter `foo` differs from its targets.\n"
                    "Given shape: (1,)\n"
                    "Target shapes:\n"
                    "{'xdot_comp.foo': (2,), 'ydot_comp.foo': (2,), 'vdot_comp.foo': (2,)}")
        self.assertEqual(expected, str(e.exception))

    def test_invalid_params_different_target_shapes_introspection_failure(self):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3),
                         ode_init_kwargs={'foo_shapes': {'xdot_comp': (1,),
                                                         'ydot_comp': (2,),
                                                         'vdot_comp': (3,)}})

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        with self.assertRaises(RuntimeError) as e:
            p.setup(check=True)

        expected = ('Invalid targets for parameter `foo`.\n'
                    'Targets have multiple shapes.\n'
                    "{'xdot_comp.foo': (1,), 'ydot_comp.foo': (2,), 'vdot_comp.foo': (3,)}")
        self.assertEqual(expected, str(e.exception))


@use_tempdirs
class TestParameterUnits(unittest.TestCase):

    def test_valid_parameters(self):
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3))

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            units='lbm',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.

        phase.set_time_val(initial=0.0, duration=2.0, units='s')
        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 9.9], units='m/s')
        phase.set_control_val('theta', [90, 90], units='deg')
        phase.set_parameter_val('foo', 5.0)

        # Run the driver to solve the problem
        p.run_driver()

        expected = convert_units(5.0, 'lbm', 'kg')

        self.assertEqual((1,), phase.parameter_options['foo']['shape'])
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.parameter_vals:foo')[-1], 5.0, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.xdot_comp.foo'), expected*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.ydot_comp.foo'), expected*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.vdot_comp.foo'), expected*np.ones(10,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.xdot_comp.foo'), expected*np.ones(20,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.ydot_comp.foo'), expected*np.ones(20,), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.vdot_comp.foo'), expected*np.ones(20,), tolerance=1.0E-5)

    def test_invalid_params_different_target_units_introspection_failure(self):
        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3),
                         ode_init_kwargs={'foo_units': {'xdot_comp': 'kg',
                                                        'ydot_comp': 'lbm',
                                                        'vdot_comp': 'slug'}})

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        with self.assertRaises(RuntimeError) as e:
            p.setup(check=True)

        expected = ("Unable to automatically assign units based on targets.\n"
                    "Targets have multiple units assigned:\n"
                    "{'xdot_comp.foo': 'kg', 'ydot_comp.foo': 'lbm', 'vdot_comp.foo': 'slug'}.\n"
                    "Either promote targets and use set_input_defaults to assign common\n"
                    "units, or explicitly provide units to the variable.")

        self.assertEqual(expected, str(e.exception))


@use_tempdirs
class TestMixedStaticDynamicParameterTargets(unittest.TestCase):

    def test_all_static(self):
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3),
                         ode_init_kwargs={'foo_static': {'xdot_comp': True, 'ydot_comp': True, 'vdot_comp': True}})

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            units='lbm',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.
        phase.set_time_val(initial=0.0, duration=2.0, units='s')
        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 9.9], units='m/s')
        phase.set_control_val('theta', [90, 90], units='deg')
        phase.set_parameter_val('foo', 5.0)

        # Run the driver to solve the problem
        p.run_driver()

        expected = convert_units(5.0, 'lbm', 'kg')

        self.assertEqual((1,), phase.parameter_options['foo']['shape'])
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.parameter_vals:foo')[-1], 5.0, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.xdot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.ydot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.vdot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.xdot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.ydot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.vdot_comp.foo'), expected, tolerance=1.0E-5)

    def test_mixed_static(self):
        import numpy as np
        import openmdao.api as om
        import dymos as dm

        #
        # Define the OpenMDAO problem
        #
        p = om.Problem(model=om.Group())

        #
        # Define a Trajectory object
        #
        traj = dm.Trajectory()

        p.model.add_subsystem('traj', subsys=traj)

        #
        # Define a Dymos Phase object with GaussLobatto Transcription
        #
        phase = dm.Phase(ode_class=_TestEOM,
                         transcription=dm.GaussLobatto(num_segments=10, order=3),
                         ode_init_kwargs={'foo_static': {'xdot_comp': True, 'ydot_comp': True, 'vdot_comp': False}})

        traj.add_phase(name='phase0', phase=phase)

        #
        # Set the time options
        # Time has no targets in our ODE.
        # We fix the initial time so that it is not a design variable in the optimization.
        # The duration of the phase is allowed to be optimized, but is bounded on [0.5, 10].
        #
        phase.set_time_options(fix_initial=True, duration_bounds=(0.5, 10.0), units='s')

        #
        # Set the time options
        # Initial values of positions and velocity are all fixed.
        # The final value of position are fixed, but the final velocity is a free variable.
        # The equations of motion are not functions of position, so 'x' and 'y' have no targets.
        # The rate source points to the output in the ODE which provides the time derivative of the
        # given state.
        phase.add_state('x', fix_initial=True, fix_final=True, units='m',
                        rate_source='xdot_comp.xdot')
        phase.add_state('y', fix_initial=True, fix_final=True, units='m',
                        rate_source='ydot_comp.ydot')
        phase.add_state('v', fix_initial=True, fix_final=False, units='m/s',
                        rate_source='vdot_comp.vdot', targets=['xdot_comp.v', 'ydot_comp.v'])

        # Define theta as a control.
        phase.add_control(name='theta', units='rad', lower=0, upper=np.pi,
                          targets=['xdot_comp.theta', 'ydot_comp.theta', 'vdot_comp.theta'])

        phase.add_parameter('foo',
                            units='lbm',
                            opt=False,
                            targets=['xdot_comp.foo', 'ydot_comp.foo', 'vdot_comp.foo'])

        # Minimize final time.
        phase.add_objective('time', loc='final')

        # Set the driver.
        p.driver = om.ScipyOptimizeDriver()

        # Allow OpenMDAO to automatically determine our sparsity pattern.
        # Doing so can significant speed up the execution of Dymos.
        p.driver.declare_coloring()

        # Setup the problem
        p.setup(check=True)

        # Now that the OpenMDAO problem is setup, we can set the values of the states.
        phase.set_time_val(initial=0.0, duration=2.0, units='s')
        phase.set_state_val('x', [0, 10], units='m')
        phase.set_state_val('y', [10, 5], units='m')
        phase.set_state_val('v', [0, 9.9], units='m/s')
        phase.set_control_val('theta', [90, 90], units='deg')
        phase.set_parameter_val('foo', 5.0)

        # Run the driver to solve the problem
        p.run_driver()

        expected = convert_units(5.0, 'lbm', 'kg')

        self.assertEqual((1,), phase.parameter_options['foo']['shape'])
        self.assertEqual({'xdot_comp.foo', 'ydot_comp.foo'}, phase.parameter_options['foo']['static_targets'])
        assert_near_equal(p.get_val('traj.phase0.timeseries.time')[-1], 1.8016, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.parameter_vals:foo')[-1], 5.0, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.xdot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.ydot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_col.vdot_comp.foo'), expected*np.ones(10), tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.xdot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.ydot_comp.foo'), expected, tolerance=1.0E-5)
        assert_near_equal(p.get_val('traj.phase0.rhs_disc.vdot_comp.foo'), expected*np.ones(20), tolerance=1.0E-5)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()



import numpy as np

import openmdao.api as om

from ..._options import options as dymos_options


class TauComp(om.ExplicitComponent):
    """
    Component that computes the phase tau, segment tau, and current segment index based on time.

    Note that stau is differentiable within a segment but non-differentiable at the segment bounds.

    Parameters
    ----------
    grid_data : GridData
        The GridData object which provides the dicretization of the Phase.
    **kwargs : dict
        Dictionary of optional arguments.
    """
    def __init__(self, grid_data=None, **kwargs):
        super().__init__(**kwargs)
        self._grid_data = grid_data
        self._no_check_partials = not dymos_options['include_check_partials']

    def initialize(self):
        """
        Declare component options.
        """
        self.options.declare('vec_size', types=int, default=1, desc='number of nodes at which to compute time')
        self.options.declare('segment_index', types=int, desc='index of the current segment')
        self.options.declare('time_units', default=None, allow_none=True, types=str,
                             desc='Units of time (or the integration variable)')

    def setup(self):
        """
        I/O creation is delayed until configure so we can determine variable shape and units.
        """
        vec_size = self.options['vec_size']
        time_units = self.options['time_units']
        dstau_dt_units = None if time_units is None else f'1/{time_units}'

        self.add_input('t_initial', val=0.0, units=time_units)
        self.add_input('t_duration', val=1.0, units=time_units)
        self.add_input('t', shape=(vec_size,), units=time_units)
        self.add_output('ptau', units=None, shape=(vec_size,))
        self.add_output('stau', units=None, shape=(vec_size,))
        self.add_output('dstau_dt', units=dstau_dt_units, val=1.0)
        self.add_output('t_phase', units=time_units, shape=(vec_size,))
        # self.add_discrete_output('segment_index', val=0)

        # Setup partials
        ar = np.arange(vec_size, dtype=int)
        self.declare_partials(of='ptau', wrt='t_initial')
        self.declare_partials(of='ptau', wrt='t_duration')
        self.declare_partials(of='ptau', wrt='t', rows=ar, cols=ar)

        self.declare_partials(of='stau', wrt='t_initial')
        self.declare_partials(of='stau', wrt='t_duration')
        self.declare_partials(of='stau', wrt='t', rows=ar, cols=ar)

        self.declare_partials(of='t_phase', wrt='t', rows=ar, cols=ar, val=1.0)
        self.declare_partials(of='t_phase', wrt='t_initial', val=-1.0)

        self.declare_partials(of='dstau_dt', wrt='t_duration', val=1.0)

    def compute(self, inputs, outputs, discrete_inputs=None, discrete_outputs=None):
        """
        Compute time component outputs.

        Parameters
        ----------
        inputs : `Vector`
            `Vector` containing inputs.
        outputs : `Vector`
            `Vector` containing outputs.
        discrete_inputs : `Vector`
            `Vector` containing discrete inputs.
        discrete_outputs : `Vector`
            `Vector` containing discrete outputs.
        """
        gd = self._grid_data
        seg_idx = self.options['segment_index']

        time = inputs['t']
        t_initial = inputs['t_initial'].copy()
        t_duration = inputs['t_duration'].copy()

        outputs['ptau'] = ptau = 2.0 * (time - t_initial) / t_duration - 1.0

        ptau0_seg = gd.segment_ends[seg_idx]
        ptauf_seg = gd.segment_ends[seg_idx + 1]

        td_seg = ptauf_seg - ptau0_seg

        outputs['stau'] = 2.0 * (ptau - ptau0_seg) / td_seg - 1.0
        outputs['dstau_dt'] = 4 / (t_duration * td_seg)
        outputs['t_phase'] = time - t_initial

    def compute_partials(self, inputs, partials):
        """
        Compute sub-jacobian parts. The model is assumed to be in an unscaled state.

        Parameters
        ----------
        inputs : Vector
            Unscaled, dimensional input variables read via inputs[key].
        partials : Jacobian
            Subjac components written to partials[output_name, input_name].
        """
        gd = self._grid_data
        seg_idx = self.options['segment_index']

        time = inputs['t']
        t_initial = inputs['t_initial']
        t_duration = inputs['t_duration']

        ptau0_seg = gd.segment_ends[seg_idx]
        ptauf_seg = gd.segment_ends[seg_idx + 1]

        td_seg = ptauf_seg - ptau0_seg

        partials['ptau', 't'] = 2.0 / t_duration
        partials['ptau', 't_initial'] = -2.0 / t_duration
        partials['ptau', 't_duration'] = -2.0 * (time - t_initial) / (t_duration ** 2)

        dstau_dptau = 2.0 / td_seg

        # Note that these derivatives ignore the effect of changing segments.
        # The derivatives of stau are discontinuous at segment boundaries.
        partials['stau', 't'] = dstau_dptau * partials['ptau', 't']
        partials['stau', 't_initial'] = dstau_dptau * partials['ptau', 't_initial']
        partials['stau', 't_duration'] = dstau_dptau * partials['ptau', 't_duration']

        partials['dstau_dt', 't_duration'] = -4 / (t_duration**2 * td_seg)



import unittest

from openmdao.utils.assert_utils import assert_near_equal
from openmdao.utils.testing_utils import use_tempdirs, require_pyoptsparse

from dymos.examples.finite_burn_orbit_raise.finite_burn_orbit_raise_problem import two_burn_orbit_raise_problem


@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaise(unittest.TestCase):

    def test_ex_two_burn_orbit_raise(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False)

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[-1], 0.3995,
                              tolerance=2.0E-3)


# This test is separate because connected phases aren't directly parallelizable.
@require_pyoptsparse(optimizer='IPOPT')
@use_tempdirs
class TestExampleTwoBurnOrbitRaiseConnected(unittest.TestCase):

    def test_ex_two_burn_orbit_raise_connected(self):
        optimizer = 'IPOPT'

        p = two_burn_orbit_raise_problem(transcription='gauss-lobatto', transcription_order=3,
                                         compressed=False, optimizer=optimizer,
                                         show_output=False, connected=True, run_driver=True)

        if p.model.traj.phases.burn2 in p.model.traj.phases._subsystems_myproc:
            assert_near_equal(p.get_val('traj.burn2.states:deltav')[0], 0.3995,
                              tolerance=4.0E-3)


if __name__ == '__main__':  # pragma: no cover
    unittest.main()
